import json
import urllib

import requests
import time

from nsj_gcf_utils.app_logger import logger
from typing import Dict, Any


class HttpUtilGetException(Exception):
    pass


class HttpUtilPostException(Exception):
    pass

class HttpUtilPutException(Exception):
    pass

class HttpUtilDeleteException(Exception):
    pass


class ResourceNotFound(Exception):
    pass


class HttpUtil:
    @staticmethod
    def post_retry(url: str, data: str, headers: Dict[str, str] = None, timeout: int = 20, tries: int = 3, interval: int = 3, format_data: bool = True, raise_for_status: bool = True, resouce_description: str = '', verify :bool = True, query_params: Dict[str, Any] = None):
        # Encoding query params
        if query_params is not None:
            q_params = urllib.parse.urlencode(query_params, doseq=True)

            if '?' in url:
                url += '&' + q_params
            else:
                url += '?' + q_params
                
        logger.info(f'Post URL: {url}')
        
        # Formatting data
        if format_data:
            if isinstance(data, dict) or isinstance(data, list):
                logger.info(
                    'Converting data from "dict" or "list" to json string')
                data = json.dumps(data)

                # adding content-type: application/json in headers
                if headers:
                    if not ('content-type' in [k.lower() for k in headers.keys()]):
                        headers['content-type'] = 'application/json'
                else:
                    headers = {'content-type': 'application/json'}
            elif not isinstance(data, str):
                logger.info('Converting data from non string to string')
                data = str(data)

        logger.info(f'Posting data (bellow) to URL: {url}')
        if not('password' in data) and not('client_secret' in data) and not('pass' in data):
            logger.info(f'Sending data: {data}')

        # Making tries
        exception_obj = None
        resp = None
        for i in range(tries):
            resp = None
            try:
                resp = requests.post(
                    url=url,
                    data=data,
                    headers=headers,
                    timeout=timeout,
                    verify=verify
                )

                if resp.status_code >= 200 and resp.status_code < 400:
                    return resp
                else:
                    logger.warning(
                        f'HTTP POST FAIL - URL: {url} - STATUS CODE {resp.status_code} - CONTENT {resp.text}')
            except Exception as e:
                logger.warning(
                    f'HTTP POST FAIL - URL: {url} - EXCEPTION {e}')
                exception_obj = e

            time.sleep(interval)
            logger.warning('Retring HTTP POST')

        # If there was no success
        if resp:
            # Checking response status
            if raise_for_status:
                if resp.status_code == 404:
                    raise ResourceNotFound(
                        f'{resouce_description} - {resp.text}')
                resp.raise_for_status()

            return resp
            # raise Exception(
            #     f'Error posting data. Status code: {resp.status_code}. Content: {resp.text}')
        else:
            raise HttpUtilPostException(
                f'Error posting data.\nMessage {exception_obj}')

    @staticmethod
    def get_retry(url: str, headers: Dict[str, str] = None, timeout: int = 20, tries: int = 3, interval: int = 3, raise_for_status: bool = True, resouce_description: str = '', data: str = None, verify :bool = True, query_params: Dict[str, Any] = None):
        logger.info(f'Get URL: {url}')
        
        # Encoding query params
        if query_params is not None:
            q_params = urllib.parse.urlencode(query_params, doseq=True)

            if '?' in url:
                url += '&' + q_params
            else:
                url += '?' + q_params
                
        logger.info(f'Get URL: {url}')

        # Making tries
        exception_obj = None
        resp = None
        for i in range(tries):
            resp = None
            try:
                resp = requests.get(
                    url=url,
                    headers=headers,
                    timeout=timeout,
                    data=data,
                    verify=verify
                )

                if resp.status_code >= 200 and resp.status_code < 400:
                    return resp
                else:
                    logger.warning(
                        f'HTTP GET FAIL - URL: {url} - STATUS CODE {resp.status_code} - CONTENT {resp.text}')
            except Exception as e:
                logger.warning(
                    f'HTTP GET FAIL - URL: {url} - EXCEPTION {e}')
                exception_obj = e

            time.sleep(interval)
            logger.warning('Retring HTTP GET')

        # If there was no success
        if resp:
            # Checking response status
            if raise_for_status:
                if resp.status_code == 404:
                    raise ResourceNotFound(
                        f'{resouce_description} - {resp.text}')
                resp.raise_for_status()

            return resp
            # raise Exception(
            #     f'Error getting data. Status code: {resp.status_code}. Content: {resp.text}')
        else:
            raise HttpUtilGetException(
                f'Error getting data.\nMessage {exception_obj}')

    @staticmethod
    def put_retry(url: str, data: str, headers: Dict[str, str] = None, timeout: int = 20, tries: int = 3,
                   interval: int = 3, format_data: bool = True, raise_for_status: bool = True,
                   resouce_description: str = '', verify :bool = True):
        # Formatting data
        if format_data:
            if isinstance(data, dict) or isinstance(data, list):
                logger.info(
                    'Converting data from "dict" or "list" to json string')
                data = json.dumps(data)

                # adding content-type: application/json in headers
                if headers:
                    if not ('content-type' in [k.lower() for k in headers.keys()]):
                        headers['content-type'] = 'application/json'
                else:
                    headers = {'content-type': 'application/json'}
            elif not isinstance(data, str):
                logger.info('Converting data from non string to string')
                data = str(data)

        logger.info(f'Puting data (bellow) to URL: {url}')
        if not ('password' in data) and not ('client_secret' in data) and not ('pass' in data):
            logger.info(f'Sending data: {data}')

        # Making tries
        exception_obj = None
        resp = None
        for i in range(tries):
            resp = None
            try:
                resp = requests.put(
                    url=url,
                    data=data,
                    headers=headers,
                    timeout=timeout,
                    verify=verify
                )

                if resp.status_code >= 200 and resp.status_code < 400:
                    return resp
                else:
                    logger.warning(
                        f'HTTP PUT FAIL - URL: {url} - STATUS CODE {resp.status_code} - CONTENT {resp.text}')
            except Exception as e:
                logger.warning(
                    f'HTTP PUT FAIL - URL: {url} - EXCEPTION {e}')
                exception_obj = e

            time.sleep(interval)
            logger.warning('Retring HTTP PUT')

        # If there was no success
        if resp:
            # Checking response status
            if raise_for_status:
                if resp.status_code == 404:
                    raise ResourceNotFound(
                        f'{resouce_description} - {resp.text}')
                resp.raise_for_status()

            return resp
        else:
            raise HttpUtilPutException(
                f'Error puting data.\nMessage {exception_obj}')

    @staticmethod
    def delete_retry(url: str, headers: Dict[str, str] = None, timeout: int = 20, tries: int = 3,
                   interval: int = 3, format_data: bool = True, raise_for_status: bool = True,
                   resouce_description: str = '', verify :bool = True):



        # Making tries
        exception_obj = None
        resp = None
        for i in range(tries):
            resp = None
            try:
                resp = requests.delete(
                    url=url,
                    headers=headers,
                    timeout=timeout,
                    verify=verify
                )

                if resp.status_code >= 200 and resp.status_code < 400:
                    return resp
                else:
                    logger.warning(
                        f'HTTP DELETE FAIL - URL: {url} - STATUS CODE {resp.status_code} - CONTENT {resp.text}')
            except Exception as e:
                logger.warning(
                    f'HTTP DELETE FAIL - URL: {url} - EXCEPTION {e}')
                exception_obj = e

            time.sleep(interval)
            logger.warning('Retring HTTP DELETE')

        # If there was no success
        if resp:
            # Checking response status
            if raise_for_status:
                if resp.status_code == 404:
                    raise ResourceNotFound(
                        f'{resouce_description} - {resp.text}')
                resp.raise_for_status()

            return resp
        else:
            raise HttpUtilDeleteException(
                f'Error deleting data.\nMessage {exception_obj}')


# import sqlalchemy
import uuid


class DBAdapter:

    # TODO Refatorar para uso de parâmetros nomeados

    def __init__(self, db_connection):
        self._db = db_connection
        self._transaction = None

    def begin(self):
        if self._transaction is None:
            self._transaction = self._db.begin()

    def commit(self):
        if self._transaction is not None:
            self._transaction.commit()
            self._transaction = None

    def rollback(self):
        if self._transaction is not None:
            self._transaction.rollback()
            self._transaction = None

    def in_transaction(self):
        return self._transaction is not None

    def execute(self, sql: str, parameters=None) -> int:
        """
        Executando uma instrução sql sem retorno.
        É obrigatório a passagem de uma conexão de banco no argumento self._db.

        Retorna o número de linhas afetadas pela instrução.
        """
        cur = None
        try:
            cur = self._execute(sql, parameters)

            return cur.rowcount
        finally:
            if cur is not None:
                cur.close()

    def execute_query_to_model(self, sql: str, model_class: object, parameters=None) -> list:
        """
        Executando uma instrução sql com retorno.
        O retorno é feito em forma de uma lista (list), com elementos do tipo passado pelo parâmetro
        "model_class".
        É importante destacar que para cada coluna do retorno, será procurado um atributo no model_class
        com mesmo nome, para setar o valor. Se este não for encontrado, a coluna do retorno é ignorada.
        """

        result = []
        cur = None
        try:
            cur = self._execute(sql, parameters)
            rs = cur.fetchall()

            for rec in rs:
                model = model_class()

                i = 0
                for column in cur.keys():
                    if (hasattr(model, column)):
                        setattr(model, column, rec[i])

                    i += 1

                result.append(model)

        finally:
            if cur is not None:
                cur.close()

        return result

    def execute_query(self, sql: str, parameters=None) -> list:
        """
        Executando uma instrução sql com retorno.
        O retorno é feito em forma de uma lista (list), com elementos do tipo dict (onde cada chave é igual ao
        nome do campo correspondente).
        """
        cur = None
        try:
            cur = self._execute(sql, parameters)
            rs = cur.fetchall()

            return [dict(rec.items()) for rec in rs]
        finally:
            if cur is not None:
                cur.close()

    def execute_query_first_result(self, sql: str, model_class: object, parameters=None) -> "model_class":
        """
        Executando uma instrução sql com retorno.
        O retorno é feito em forma de um objeto do tipo passado pelo parâmetro "model_class".
        É importante destacar que para cada coluna do retorno, será procurado um atributo no model_class
        com mesmo nome, para setar o valor. Se este não for encontrado, a coluna do retorno é ignorada.
        """

        result = None
        cur = None
        try:
            cur = self._execute(sql, parameters)
            rs = cur.fetchone()

            if (len(rs) > 0):
                model = model_class()

                i = 0
                for column in cur.keys():
                    if (hasattr(model, column)):
                        setattr(model, column, rs[i])

                    i += 1

                result = model

        finally:
            if cur is not None:
                cur.close()

        return result

    def get_single_result(self, sql: str, parameters=None):
        """
        Executa uma instrução SQL para a qual se espera um único retorno (com tipo primitivo). Exemplo:
        select 1+1
        Se não houver retorno, retorna None.
        """
        cur = None
        try:
            cur = self._execute(sql, parameters)
            return cur.scalar()
        finally:
            if cur is not None:
                cur.close()

    def _check_type(self, parameter):
        if (isinstance(parameter, uuid.UUID)):
            return str(parameter)
        else:
            return parameter

    def _execute(self, sql: str, parameters: list):
        new_transaction = not self.in_transaction()

        try:
            if new_transaction:
                self.begin()

            if (parameters != None):
                pars = [self._check_type(par) for par in parameters]
                return self._db.execute(sql, pars)
            else:
                return self._db.execute(sql)
        except:
            if new_transaction:
                self.rollback()
            raise
        finally:
            if new_transaction:
                self.commit()

    def execute_query_from_file(self, query_file_path, *parameters):
        with open(query_file_path) as f:
            sql = f.read()
        return self.execute_query(sql, parameters)


# class AlchemyDBAdapter:
#     def __init__(self, db_connection):
#         self._db = db_connection

#     def execute_query(self, sql: str, **kwargs) -> list:
#         """
#         Executando uma instrução sql com retorno.
#         O retorno é feito em forma de uma lista (list), com elementos do tipo dict (onde cada chave é igual ao
#         nome do campo correspondente).
#         """
#         cur = None
#         sql = sqlalchemy.text(sql)
#         try:
#             cur = self._db.execute(sql, **kwargs)
#             rs = cur.fetchall()

#             return [dict(rec.items()) for rec in rs]
#         finally:
#             if cur is not None:
#                 cur.close()

#     def execute(self, sql: str, **kwargs) -> int:
#         """
#         Executando uma instrução sql sem retorno.
#         É obrigatório a passagem de uma conexão de banco no argumento self._db.

#         Retorna o número de linhas afetadas pela instrução.
#         """
#         cur = None
#         sql = sqlalchemy.text(sql)
#         try:
#             with self._db.begin():
#                 cur = self._db.execute(sql, **kwargs)

#             return cur.rowcount
#         finally:
#             if cur is not None:
#                 cur.close()

#     def execute_query_from_file(self, query_file_path, **kwargs):
#         with open(query_file_path) as f:
#             sql = f.read()
#         return self.execute_query(sql, **kwargs)

#     def execute_from_file(self, sql_file_path, **kwargs):
#         with open(sql_file_path) as f:
#             sql = f.read()
#         return self.execute(sql, **kwargs)



import base64
import re
import requests


class NsjAuthenticationService:

    def __init__(self, client_id: str, client_secret: str, introspection_url: str) -> None:
        self._client_id = client_id
        self._client_secret = client_secret
        self._introspection_url = introspection_url

    def authenticate(self, access_token: str) -> str:
        """
        Check if access_token is valid (throught token introspection endpoint),
        and returns access_token info, or None (is not valid).
        """

        # Checking heaaders
        if not access_token:
            return None

        # Checking format (throught regex)
        matcher = re.compile('^Bearer (.+)$')
        match = matcher.match(access_token)
        if not match:
            return None
        access_token = match.group(1)

        # Calling Introspection Endpoint
        basic = f'{self._client_id}:{self._client_secret}'
        basic = base64.b64encode(basic.encode('utf-8')).decode('utf-8')
        basic = f'Basic {basic}'

        data = {
            'token': access_token,
            'token_type_hint': 'access_token'
        }

        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': basic
        }

        resp = requests.post(url=self._introspection_url,
                             data=data, headers=headers, timeout=60)

        if resp.status_code == 401:
            return Exception('Authentication error in token introspection. Check client_id and client_secret.')

        resp.raise_for_status()
        resp = resp.json()

        # Testing result
        if not ('username' in resp):
            return None

        # Returning token info
        return resp



import uuid

from sqlparams import SQLParams


class DBAdapter2:

    def __init__(self, db_connection):
        self._db = db_connection
        self._transaction = None

    def begin(self):
        if self._transaction is None:
            self._transaction = self._db.begin()

    def commit(self):
        if self._transaction is not None:
            self._transaction.commit()
            self._transaction = None

    def rollback(self):
        if self._transaction is not None:
            self._transaction.rollback()
            self._transaction = None

    def in_transaction(self):
        return self._transaction is not None

    def execute(self, sql: str, **kwargs) -> int:
        """
        Executando uma instrução sql, com ou sem retorno.
        É obrigatório a passagem de uma conexão de banco no argumento self._db.

        Retorna o número de linhas afetadas pela instrução e o retorno esperado se houver, como uma tupla.
        O retorno é feito em forma de uma lista (list), com elementos do tipo dict (onde cada chave é igual ao
        nome do campo correspondente).
        """
        cur = None
        try:
            cur = self._execute(sql, **kwargs)

            if 'returning' in sql.lower():
                rs = cur.fetchall()
                returning = [dict(rec.items()) for rec in rs]
            else:
                returning = None

            return (cur.rowcount, returning)
        finally:
            if cur is not None:
                cur.close()

    def execute_query_to_model(self, sql: str, model_class: object, **kwargs) -> list:
        """
        Executando uma instrução sql com retorno.
        O retorno é feito em forma de uma lista (list), com elementos do tipo passado pelo parâmetro
        "model_class".
        É importante destacar que para cada coluna do retorno, será procurado um atributo no model_class
        com mesmo nome, para setar o valor. Se este não for encontrado, a coluna do retorno é ignorada.
        """

        result = []
        cur = None
        try:
            cur = self._execute(sql, **kwargs)
            rs = cur.fetchall()

            for rec in rs:
                model = model_class()

                i = 0
                for column in cur.keys():
                    if (hasattr(model, column)):
                        setattr(model, column, rec[i])

                    i += 1

                result.append(model)

        finally:
            if cur is not None:
                cur.close()

        return result

    def execute_query(self, sql: str, **kwargs) -> list:
        """
        Executando uma instrução sql com retorno.
        O retorno é feito em forma de uma lista (list), com elementos do tipo dict (onde cada chave é igual ao
        nome do campo correspondente).
        """
        cur = None
        try:
            cur = self._execute(sql, **kwargs)
            rs = cur.fetchall()

            return [dict(rec.items()) for rec in rs]
        finally:
            if cur is not None:
                cur.close()

    def execute_query_first_result(self, sql: str, **kwargs) -> list:
        """
        Executando uma instrução sql com retorno.
        O retorno é feito em forma de um dict (onde cada chave é igual ao nome do campo correspondente).

        Apenas o primeiro registro é retornado (os demais serão descartados, se houverem).
        Caso não haja registros correspondentes a query, retorna None.
        """
        cur = None
        try:
            cur = self._execute(sql, **kwargs)
            rs = cur.fetchall()

            results = [dict(rec.items()) for rec in rs]

            if len(results) <= 0:
                return None
            else:
                return results[0]
        finally:
            if cur is not None:
                cur.close()

    def execute_query_first_result_to_model(self, sql: str, model_class: object, **kwargs) -> "model_class":
        """
        Executando uma instrução sql com retorno.
        O retorno é feito em forma de um objeto do tipo passado pelo parâmetro "model_class".
        É importante destacar que para cada coluna do retorno, será procurado um atributo no model_class
        com mesmo nome, para setar o valor. Se este não for encontrado, a coluna do retorno é ignorada.
        """

        result = None
        cur = None
        try:
            cur = self._execute(sql, **kwargs)
            rs = cur.fetchone()

            if (len(rs) > 0):
                model = model_class()

                i = 0
                for column in cur.keys():
                    if (hasattr(model, column)):
                        setattr(model, column, rs[i])

                    i += 1

                result = model

        finally:
            if cur is not None:
                cur.close()

        return result

    def get_single_result(self, sql: str, **kwargs):
        """
        Executa uma instrução SQL para a qual se espera um único retorno (com tipo primitivo). Exemplo:
        select 1+1
        Se não houver retorno, retorna None.
        """
        cur = None
        try:
            cur = self._execute(sql, **kwargs)
            return cur.scalar()
        finally:
            if cur is not None:
                cur.close()

    def _check_type(self, parameter):
        if (isinstance(parameter, uuid.UUID)):
            return str(parameter)
        else:
            return parameter

    def _execute(self, sql: str, **kwargs):

        new_transaction = not self.in_transaction()

        try:
            if new_transaction:
                self.begin()

            if (kwargs is not None):
                pars = {key: self._check_type(kwargs[key]) for key in kwargs}
                sql2, pars2 = SQLParams('named', 'format').format(sql, pars)
                return self._db.execute(sql2, pars2)
            else:
                return self._db.execute(sql)
        except:
            if new_transaction:
                self.rollback()
            raise

        finally:
            if new_transaction:
                self.commit()

    def execute_query_from_file(self, query_file_path, **kwargs):
        with open(query_file_path, 'r') as f:
            sql = f.read()
        return self.execute_query(sql, **kwargs)



import urllib.parse
import uuid

from nsj_gcf_utils.exception import PaginationException
from typing import Any, List


def page_body(
    base_url: str,
    limit: int,
    current_after: uuid.UUID,
    current_before: uuid.UUID,
    result: List[Any],
    id_field: str = 'id'
):

    if current_after is not None and current_before is None:
        return _page_body_after(
            base_url=base_url,
            limit=limit,
            current_after=current_after,
            result=result,
            id_field=id_field
        )
    # elif current_after is None and current_before is not None:
    #     return _page_body_before(
    #         base_url=base_url,
    #         limit=limit,
    #         current_before=current_before,
    #         result=result,
    #         id_field=id_field
    #     )
    elif current_after is not None and current_before is not None:
        raise PaginationException(
            'Não é permitido usar os parâmetros "after" (ou "offset") e "before" simultâneamente.')

    # Checking has next and previous
    has_next = len(result) >= limit

    # Preparing previous and next URLs
    url = base_url
    if '?' in base_url:
        url += '&'
    else:
        url += '?'

    last = -1 if limit >= len(result) else limit-1
    if has_next:
        params_next = {
            'after': result[last][id_field],
            'limit': limit
        }
        params_next = urllib.parse.urlencode(params_next, doseq=True)
        url_next = url + params_next
    else:
        url_next = None

    url_previous = None

    # Returning pagination body
    return {
        'next': url_next,
        # 'prev': url_previous,
        'result': result[0:limit]
    }


def _page_body_after(
    base_url: str,
    limit: int,
    current_after: uuid.UUID,
    result: List[Any],
    id_field: str = 'id'
):
    # Checking has next and previous
    has_next = len(result) >= limit
    has_prev = True

    # Preparing previous and next URLs
    url = base_url
    if '?' in base_url:
        url += '&'
    else:
        url += '?'

    last = -1 if limit >= len(result) else limit-1
    if has_next:
        params_next = {
            'after': result[last][id_field],
            'limit': limit
        }
        params_next = urllib.parse.urlencode(params_next, doseq=True)
        url_next = url + params_next
    else:
        url_next = None

    if has_prev:
        params_prev = {
            'before': current_after,
            'limit': limit
        }
        params_prev = urllib.parse.urlencode(params_prev, doseq=True)
        url_previous = url + params_prev
    else:
        url_previous = None

    # Returning pagination body
    return {
        'next': url_next,
        # 'prev': url_previous,
        'result': result[0:limit]
    }


def _page_body_before(
    base_url: str,
    limit: int,
    current_before: uuid.UUID,
    result: List[Any],
    id_field: str = 'id'
):
    # Checking has next and previous
    has_next = True
    has_prev = len(result) >= limit

    # Preparing previous and next URLs
    url = base_url
    if '?' in base_url:
        url += '&'
    else:
        url += '?'

    if has_next:
        params_next = {
            'after': current_before,
            'limit': limit
        }
        params_next = urllib.parse.urlencode(params_next, doseq=True)
        url_next = url + params_next
    else:
        url_next = None

    first = 0 if limit > len(result) else -1*limit
    if has_prev:
        params_prev = {
            'before': result[first][id_field],
            'limit': limit
        }
        params_prev = urllib.parse.urlencode(params_prev, doseq=True)
        url_previous = url + params_prev
    else:
        url_previous = None

    # Returning pagination body
    return {
        'next': url_next,
        'prev': url_previous,
        'result': result[first:]
    }



import logging
import os
import sys

log_format = logging.Formatter(
    '%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S')

app_name = os.getenv('APP_NAME', 'MISSING_APP_NAME_ENV_VAR')
logger = logging.getLogger(app_name)
logger.setLevel(logging.DEBUG)

consoleHandler = logging.StreamHandler(sys.stdout)
consoleHandler.setFormatter(log_format)
logger.addHandler(consoleHandler)



from hashlib import sha256
from nsj_gcf_utils.json_util import json_dumps

"""Cria um hash usando SHA256 para ser usado para autenticação pelo webhook.
   Ele recebe a url, o body e o metodo da requisição junto com uma chave especial.
   Depois junta todos os dados e usa SHA256 para criar o hash.
"""
def hash_webhook(url, body, method, key):
    # Converte o body para string caso o body seja um dict(json)
    if isinstance(body, dict):
        body = json_dumps(body)

    # tipagem para ajudar o autocomplete de IDEs
    # str na key caso ela seja um número ou um uuid
    hash: str = url + body + method + str(key) 

    return sha256(hash.encode()).hexdigest()

# Teste para o integrador vmpay
if __name__ == '__main__':
    print(hash_webhook("http://localhost/produtos",{
        "esquema": "estoque",
        "tabela": "produtos",
        "operacao": "I",
        "dados": {
            "especificacao": "teste14",
            "codigo": "1234",
            "codigodebarras": "",
            "estabelecimento":"5e5a86b7-d494-4c31-806a-a5e74862403a",
            "grupoempresarial": "0ef4df5d-9881-4c3c-b8ea-a09149f799a5",
            "grupodeinventario": 0
        }
    }, "POST","124325"))


from typing import Tuple

from nsj_gcf_utils.http_util import HttpUtil


class KeycloackService:
    """
    Class to authentication with KeyCloak.
    """
    _keycloack_token_url: str

    def __init__(self, keycloack_token_url: str) -> None:
        self._keycloack_token_url = keycloack_token_url

    # TODO Melhorar controle de conexão, incluindo renovação do acess_token pelo refresh_token

    def authenticate(self, client_id: str, username: str, password: str, scope: str, client_secret: str = None) -> Tuple[str, int, str, int]:
        """
        Authenticate with keycloak.

        Returns a tuple, with:
        [access_token: str, expires_in: int, refresh_token: str, refresh_expires_in: int]
        """

        # Making data
        headers = {'Content-type': 'application/x-www-form-urlencoded'}
        data = {
            'client_id': client_id,
            'scope': scope,
            'grant_type': 'password',
            'username': username,
            'password': password
        }

        if client_secret is not None:
            data['client_secret'] = client_secret

        # Calling API
        resp = HttpUtil.post_retry(self._keycloack_token_url,
                                   headers=headers, data=data, format_data=False, resouce_description='get_access_token')

        # Reading result
        resp = resp.json()

        return (
            resp['access_token'],
            resp['expires_in'],
            resp['refresh_token'],
            resp['refresh_expires_in']
        )



class IBAN:
    def __init__(self):
        pass

    def setPais(self, pais:str):
        if pais != "BR":
            raise Exception("Pais não suportado")
        self.pais = pais
        return self

    def getPais(self)->str:
        return self.pais

    def setBanco(self, ispb:str):
        self.banco = ispb.zfill(8)
        return self

    def getBanco(self)->str:
        return self.banco

    def setAgencia(self, agencia:str):
        self.agencia = agencia.zfill(5)
        return self

    def getAgencia(self)->str:
        return self.agencia

    def setConta(self, conta:str):
        self.conta = conta.zfill(10)
        return self

    def getConta(self)->str:
        return self.conta

    def setTipo(self, tipo:str):
        if not tipo in ['C', 'c', 'P', 'p']:
            raise Exception("O tipo da conta deve ser C para Conta Corrente ou P para Conta Poupança")
        self.tipo = tipo
        return self

    def getTipo(self)->str:
        return self.tipo

    def setTitular(self, titular:int):
        if titular > 0 and titular < 10:
            self.titular = chr(titular + ord('0'))
        elif titular >= 10 and titular <= 35:
            self.titular = chr(titular - 9 + ord('A'))
        else:
            raise Exception("O número do titular deve ser um inteiro entre 1 e 35")
        return self

    def getTitular(self)->int:
        if self.titular >= '1' and self.titular <= '9':
            return ord(self.titular) - ord('0')
        return ord(self.titular) - ord('A')

    def getIban(self):
        return IBAN.preencherDv(self.pais + '00' + self.banco + self.agencia + self.conta + self.tipo + self.titular)

    def parseIban(self, iban:str, check:bool = False):
        if check and not IBAN.validarDv(iban):
            raise Exception("O IBAN inserido não é válido")
        self.pais = iban[:2]
        self.banco = iban[4:12]
        self.agencia = iban[12:17]
        self.conta = iban[17:27]
        self.tipo = iban[27]
        self.titular = iban[28]

    def preencherDv(iban:str):
        soma = 0;
        for c in iban[4:] + iban[:2]:
            if c >= '0' and c <= '9':
                soma = soma * 10 + ord(c) - ord('0')
            elif c >= 'A' and c <= 'Z':
                soma = soma * 100 + ord(c) - ord('A') + 10
            elif c >= 'a' and c <= 'z':
                soma = soma * 100 + ord(c) - ord('a') + 10
            else:
                raise Exception()
        return iban[:2] + str(98 - soma * 100 % 97) + iban[4:]

    def validarDv(iban:str):
        soma = 0;
        for c in iban[4:] + iban[:4]:
            if c >= '0' and c <= '9':
                soma = soma * 10 + ord(c) - ord('0')
            elif c >= 'A' and c <= 'Z':
                soma = soma * 100 + ord(c) - ord('A') + 10
            elif c >= 'a' and c <= 'z':
                soma = soma * 100 + ord(c) - ord('a') + 10
            else:
                raise
        return soma % 97 == 1
